#include <iostream>
#include <Geom_Plane.hxx>
#include <Geom_TrimmedCurve.hxx>
#include <GC_MakeSegment.hxx>
#include <TopExp.hxx>
#include <TopoDS.hxx>
#include <TopoDS_Shape.hxx>
#include <TopoDS_Edge.hxx>
#include <TopoDS_Wire.hxx>
#include <TopoDS_Face.hxx>
#include <TopExp_Explorer.hxx>
#include <TopLoc_Location.hxx>
#include <TopTools_ListIteratorOfListOfShape.hxx>
#include <TopTools_ListOfShape.hxx>
#include <TopTools_IndexedMapOfShape.hxx>
#include <TNaming_NamedShape.hxx>
#include <TNaming_UsedShapes.hxx>
#include <TNaming_Builder.hxx>
#include <TNaming_Selector.hxx>
#include <TNaming_Tool.hxx>
#include <BRep_Tool.hxx>
#include <BRepTools.hxx>
#include <BRepPrimAPI_MakeBox.hxx>
#include <BRepPrimAPI_MakePrism.hxx>
#include <BRepBuilderAPI_MakeEdge.hxx>
#include <BRepBuilderAPI_MakeWire.hxx>
#include <BRepBuilderAPI_MakeFace.hxx>
#include <BRepAlgoAPI_Cut.hxx>
#include <TDF_Data.hxx>
#include <TDF_Label.hxx>
#include <TDF_LabelMap.hxx>
#include <TDF_ChildIterator.hxx>
#include <gp_Pnt.hxx>
#include <gp_Dir.hxx>
#include <gp_Vec.hxx>
#include <gp_Ax1.hxx>
#include <gp_Dir.hxx>
#include <gp_Ax2.hxx>

void printShapeInfos(TopoDS_Shape shape1, TopoDS_Shape shape2, TopoDS_Shape shape3);
void writeShapeInfos(TopoDS_Shape shape1, TopoDS_Shape shape2, TopoDS_Shape shape3);
void printShapeInfo(TopoDS_Shape shape);
void writeShapeInfo(TopoDS_Shape shape, std::ofstream&);
void printDumpFile(TopoDS_Shape shape);
void writeDumpFiles(TopoDS_Shape orig, TopoDS_Shape cut, TopoDS_Shape newBox);

void printShapeInfos(TopoDS_Shape shape1, TopoDS_Shape shape2, TopoDS_Shape shape3)
{
  std::cout << "orig box" << std::endl;
  printShapeInfo(shape1);
  std::cout << "cut box" << std::endl;
  printShapeInfo(shape2);
  std::cout << "new box" << std::endl;
  printShapeInfo(shape3);
}

void writeShapeInfos(TopoDS_Shape shape1, TopoDS_Shape shape2, TopoDS_Shape shape3)
{
  std::ofstream myOutFile;
  // open output files
  myOutFile.open("shapeInfos.txt");
  myOutFile << "orig box" << std::endl;
  writeShapeInfo(shape1, myOutFile);
  myOutFile << "cut box" << std::endl;
  writeShapeInfo(shape2, myOutFile);
  myOutFile << "new box" << std::endl;
  writeShapeInfo(shape3, myOutFile);
  myOutFile.close();
}

void printShapeInfo(TopoDS_Shape shape)
{
  TopTools_IndexedMapOfShape mapOfShapes;
  TopExp::MapShapes(shape, TopAbs_FACE, mapOfShapes);
  for (int i = 1; i <= mapOfShapes.Extent(); i++)
  {
    TopoDS_Face aFace = TopoDS::Face(mapOfShapes.FindKey(i));
    Handle(Geom_Surface) aSurface = BRep_Tool::Surface(aFace);
    if (aSurface->DynamicType() == STANDARD_TYPE(Geom_Plane))
    {
      Handle(Geom_Plane) aPlane = Handle(Geom_Plane)::DownCast(aSurface);
      gp_Pnt aPoint = aPlane->Location();
      gp_Ax1 anAxis = aPlane->Axis();
      gp_Dir aDir = anAxis.Direction();
      Standard_Real locx, locy, locz, dirx, diry, dirz;
      locx = aPoint.X(); locy = aPoint.Y();
      locz = aPoint.Z(); dirx = aDir.X(); diry = aDir.Y(); dirz = aDir.Z();
      std::cout << "For i= " << i << " ";
      std::cout << "loc = (" << locx << ", " << locy << ", " << locz << ") , ";
      std::cout << "dir = (" << dirx << ", " << diry << ", " << dirz << ") " << std::endl;
    }
  }
}

void writeShapeInfo(TopoDS_Shape shape, std::ofstream& myOutFile)
{
  TopTools_IndexedMapOfShape mapOfShapes;
  TopExp::MapShapes(shape, TopAbs_FACE, mapOfShapes);
  for (int i = 1; i <= mapOfShapes.Extent(); i++)
  {
    TopoDS_Face aFace = TopoDS::Face(mapOfShapes.FindKey(i));
    Handle(Geom_Surface) aSurface = BRep_Tool::Surface(aFace);
    if (aSurface->DynamicType() == STANDARD_TYPE(Geom_Plane))
    {
      Handle(Geom_Plane) aPlane = Handle(Geom_Plane)::DownCast(aSurface);
      gp_Pnt aPoint = aPlane->Location();
      gp_Ax1 anAxis = aPlane->Axis();
      gp_Dir aDir = anAxis.Direction();
      Standard_Real locx, locy, locz, dirx, diry, dirz;
      locx = aPoint.X();
      locy = aPoint.Y();
      locz = aPoint.Z();
      dirx = aDir.X();
      diry = aDir.Y();
      dirz = aDir.Z();
      myOutFile << "For i= " << i << " ";
      myOutFile << "loc = (" << locx << ", " << locy << ", " << locz << ") , ";
      myOutFile << "dir = (" << dirx << ", " << diry << ", " << dirz << ") " << std::endl;
    }
  }
}

void printDumpFile(TopoDS_Shape shape)
{
  BRepTools::Dump(shape, std::cout);
}

void writeDumpFiles(TopoDS_Shape orig, TopoDS_Shape cut, TopoDS_Shape newBox)
{
  std::ofstream myBoxFile, cutBoxFile, newBoxFile;
  // open output files
  myBoxFile.open("shape1.txt");
  cutBoxFile.open("shape2.txt");
  newBoxFile.open("shape3.txt");
  BRepTools::Dump(orig, myBoxFile);
  BRepTools::Dump(cut, cutBoxFile);
  BRepTools::Dump(newBox, newBoxFile);
  // close output files
  myBoxFile.close();
  cutBoxFile.close();
  newBoxFile.close();
}

void cutBoxAndShowThings(TopoDS_Shape origBox)
{
  TopoDS_Shape cutBox, newBox;
  gp_Pnt cut_loc; gp_Dir cut_dir; gp_Ax2 cut_ax2;
  // create box to cut out and write it out.
  cut_loc = gp_Pnt(7, 5, 30);
  cut_dir = gp_Dir(0, 0, -1);
  cut_ax2 = gp_Ax2(cut_loc, cut_dir);
  cutBox = BRepPrimAPI_MakeBox(cut_ax2, 5., 5., 5.).Shape();
  // create new box by cutting out the second box from the first
  BRepAlgoAPI_Cut builder(origBox, cutBox);
  newBox = builder.Shape();
  // print out Dump to files
  writeDumpFiles(origBox, cutBox, newBox);
  //printDumpFile(origBox);
  //printDumpFile(cutBox);
  //printDumpFile(newBox);
  writeShapeInfos(origBox, cutBox, newBox);
  //printShapeInfos(origBox, cutBox, newBox);
}

void runcase3()
{
    TopoDS_Shape box2, slot2, cutBox2, fillet2;
    gp_Pnt cut_loc(9.0, 4.0, 0.0);

    Handle(TDF_Data) myDataFramework = new TDF_Data();
    TDF_Label Root            = myDataFramework->Root();
    TDF_Label box2Label = TDF_TagSource::NewChild(Root);
    TDF_Label slot2Label = TDF_TagSource::NewChild(Root);
    TDF_Label selection2Label  = TDF_TagSource::NewChild(Root);
    TDF_Label cutBox2Label = TDF_TagSource::NewChild(Root);

    int count = 0, countDel = 0;
    TopTools_IndexedMapOfShape Edges;
    TDF_ChildIterator itchild;

    box2 = BRepPrimAPI_MakeBox(10.0, 10.0, 10.0).Shape();
    TNaming_Builder box2Builder(box2Label);
    box2Builder.Generated(box2);
    TopExp_Explorer ExBox2(box2, TopAbs_FACE);
    for (; ExBox2.More(); ExBox2.Next()) {
        TDF_Label faceLabel = TDF_TagSource::NewChild(box2Label);
        TNaming_Builder faceBuilder(faceLabel);
        faceBuilder.Generated(ExBox2.Current());
    }

    TNaming_Selector selector2(selection2Label);
    Edges.Clear();
    TopExp::MapShapes(box2, TopAbs_EDGE, Edges);
    // from trial and error it was found we want this edge
    TopoDS_Shape fillet2Edge = Edges.FindKey(6);
    selector2.Select(fillet2Edge, box2);

    slot2 = BRepPrimAPI_MakeBox(cut_loc, 1.0, 2.0, 10.0).Shape();
    TNaming_Builder slot2Builder(slot2Label);
    slot2Builder.Generated(slot2);
    TopExp_Explorer ExSlot2(slot2, TopAbs_FACE);
    for (; ExSlot2.More(); ExSlot2.Next()) {
        TDF_Label faceLabel = TDF_TagSource::NewChild(slot2Label);
        TNaming_Builder faceBuilder(faceLabel);
        faceBuilder.Generated(ExSlot2.Current());
    }

    BRepAlgoAPI_Cut cutBox2Maker(box2, slot2);
    cutBox2 = cutBox2Maker.Shape();
    TNaming_Builder cutBox2Builder(cutBox2Label);
    cutBox2Builder.Modify(box2, cutBox2);
//    cutBox2Builder.Modify(slot2, cutBox2);
    count = 0; countDel = 0;
    TDF_Label modif2FacesLabelFromBox2 = TDF_TagSource::NewChild(cutBox2Label);
    TDF_Label del2FacesLabelFromBox2 = TDF_TagSource::NewChild(cutBox2Label);
    TDF_Label modif2FacesLabelFromSlot2 = TDF_TagSource::NewChild(cutBox2Label);
    TDF_Label del2FacesLabelFromSlot2 = TDF_TagSource::NewChild(cutBox2Label);
    TNaming_Builder modif2FacesBuilderFromBox2(modif2FacesLabelFromBox2);
    TNaming_Builder del2FacesBuilderFromBox2(del2FacesLabelFromBox2);
    TNaming_Builder modif2FacesBuilderFromSlot2(modif2FacesLabelFromSlot2);
    TNaming_Builder del2FacesBuilderFromSlot2(del2FacesLabelFromSlot2);
    ExBox2.ReInit();
    for (; ExBox2.More(); ExBox2.Next()) {
        TopoDS_Shape oldShape = ExBox2.Current(), newShape;
        if (cutBox2Maker.IsDeleted(oldShape)) {
            del2FacesBuilderFromBox2.Delete(oldShape);
            countDel++;
            cout << count << ", " << countDel << endl;
            continue;
        }
        for(TopTools_ListIteratorOfListOfShape it(cutBox2Maker.Modified(oldShape));
                it.More(); it.Next()) {
            newShape = it.Value();
            if (!oldShape.IsSame(newShape)) {
                modif2FacesBuilderFromBox2.Modify(oldShape, newShape);
                count++;
            }
        }
        cout << count << ", " << countDel << endl;
    }
    ExSlot2.ReInit();
    for (; ExSlot2.More(); ExSlot2.Next()) {
        TopoDS_Shape oldShape = ExSlot2.Current(), newShape;
        if (cutBox2Maker.IsDeleted(oldShape)) {
            del2FacesBuilderFromSlot2.Delete(oldShape);
            countDel++;
            cout << count << ", " << countDel << endl;
            continue;
        }
        // for (TopoDS_Shape modFace: cutBox1Maker.Modified(oldShape)) {
        for(TopTools_ListIteratorOfListOfShape it(cutBox2Maker.Modified(oldShape));
            it.More(); it.Next()) {
            newShape = it.Value();
            if (!oldShape.IsSame(newShape)) {
                modif2FacesBuilderFromSlot2.Modify(oldShape, newShape);
                cout << "======" << endl;
                printShapeInfo(oldShape);
                cout << "------" << endl;
                printShapeInfo(newShape);
                cout << "======" << endl;
                count++;
            }
        }
        cout << count << ", " << countDel << endl;
    }

    TDF_LabelMap scope2;
    cout << selector2.Solve(scope2) << endl;
    TopoDS_Shape targetEdge2 = TNaming_Tool::GetShape(selector2.NamedShape());

    writeDumpFiles(box2, slot2, targetEdge2);
    writeShapeInfos(box2, slot2, targetEdge2);
}

int main()
{
    runcase3();
    return 0;
}
